// mobjToTable: copies over relevant properties for carrying from b to a
// a: the mobj or table to copy values into
// b: the mobj or table to copy values from
freeslot(
	"sfx_m2jump",
	"sfx_pluck",
	"sfx_toss",
	"SPR_VEGI",
	"S_PLAY_PLUCK1",
	"S_PLAY_PLUCK2",
	"S_PLAY_CARRYCROUCH",
	"S_CARRYDISPLAY",
	"S_THROWNENEMY_SPAWN",
	"S_VEGGIE",
	"MT_CARRYDISPLAY",
	"MT_THROWNENEMY",
	"MT_VEGGIE"
)

local mobjToTable = function(a,b)
	if not (a and b
	and b.valid)
		return false end
		
	a.state = b.state
	a.sprite = b.sprite
	a.frame = b.frame
	a.anim_duration = b.anim_duration
	a.renderflags = b.renderflags
	a.blendmode = b.blendmode
	a.spritexscale = b.spritexscale
	a.spriteyscale = b.spriteyscale
	a.spritexoffset = b.spritexoffset
	a.spriteyoffset = b.spriteyoffset
	a.radius = b.radius
	a.height = b.height
	a.tics = b.tics
	a.flags = b.flags
	a.flags2 = b.flags2
	a.color = b.color
	a.colorized = b.colorized
	a.drawonlyforplayer = b.drawonlyforplayer
	a.dontdrawforviewmobj = b.dontdrawforviewmobj
	a.hnext = b.hnext
	a.hprev = b.hprev
	a.health = b.health
	a.movedir = b.movedir
	a.movecount = b.movecount
	a.target = b.target
	a.reactiontime = b.reactiontime
	a.threshold = b.threshold
	a.lastlook = b.lastlook
	a.spawnpoint = b.spawnpoint
	a.tracer = b.tracer
	a.friction = b.friction
	a.movefactor = b.movefactor
	a.fuse = b.fuse
	a.scale = b.scale
	a.destscale = b.destscale
	a.scalespeed = b.scalespeed
	a.extravalue1 = b.extravalue1
	a.extravalue2 = b.extravalue2
	a.cusval = b.cusval
	a.cvmem = b.cvmem
	a.mirrored = b.mirrored
	a.shadowscale = b.shadowscale
	a.dispoffset = b.dispoffset
	
	if b.hitfunction
		a.hitfunction = b.hitfunction
	end
	
	return true
end

local throwEnemy = function(mo,speed)
	if not mo and mo.valid
		return end
	
	if not (mo.carrying and mo.carrying.valid
	and mo.carrydisplay and mo.carrydisplay.valid)
		mo.carrying = nil
		mo.carrydisplay = nil
		return end
		
	S_StartSound(mo,sfx_toss)
	
	local thrown = P_SpawnMobj(mo.carrydisplay.x,mo.carrydisplay.y,mo.carrydisplay.z,MT_THROWNENEMY)
	if not thrown and thrown.valid
		return true end
			
	thrown.target = mo
	thrown.angle = mo.angle
	P_InstaThrust(thrown,mo.angle,speed)
	
	thrown.momx = $+mo.momx
	thrown.momy = $+mo.momy
	thrown.momz = $+mo.momz
	
	thrown.carrying = {}
	mobjToTable(thrown.carrying,mo.carrying)
	
	thrown.carrying.type = mo.carrying.type
	thrown.carrying.valid = true
	mo.carrying = nil
	
	if thrown.carrying.type == MT_NULL
		thrown.momx = $*2
		thrown.momy = $*2
	end
	
	return true
end

local A_SpawnCarriedEnemy = function(t,a,b)
	local enemy = P_SpawnMobj(t.x,t.y,t.z,t.carrying.type)
	if not (enemy and enemy.valid)
		return
	end

	mobjToTable(enemy,t.carrying)
	
	if ~enemy.flags & MF_NOGRAVITY
		enemy.momx = t.momx
		enemy.momy = t.momy
		enemy.momz = t.momz / -2
		enemy.angle = t.angle
	end
	
	if not (enemy.landfunction
	and enemy.landfunction(enemy,t,t.hitsomething))
	and t.hitsomething
		P_DamageMobj(enemy,t.target,t.target)
	end
end

states[S_PLAY_PLUCK1] = {
	sprite = SPR_PLAY,
	frame = SPR2_MLEE,
	tics = TICRATE/6,
	nextstate = S_PLAY_PLUCK2
}

states[S_PLAY_PLUCK2] = {
	sprite = SPR_PLAY,
	frame = SPR2_LAND,
	tics = TICRATE/6,
	nextstate = S_PLAY_FALL
}

states[S_PLAY_CARRYCROUCH] = {
	sprite = SPR_PLAY,
	frame = SPR2_LAND,
	tics = -1,
	nextstate = S_PLAY_FALL
}

states[S_CARRYDISPLAY] = {
	sprite = SPR_NULL,
	frame = A,
	tics = -1
}

states[S_THROWNENEMY_SPAWN] = {
	sprite = SPR_NULL,
	frame = A,
	action = A_SpawnCarriedEnemy,
	tics = 0
}

states[S_VEGGIE] = {
	sprite = SPR_VEGI,
	frame = A,
	tics = 1,
	nextstate = S_SMOKE1
}

mobjinfo[MT_CARRYDISPLAY] = {
	spawnstate = S_CARRYDISPLAY,
	spawnhealth = 1000,
	painstate = S_CARRYDISPLAY,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING
}

mobjinfo[MT_THROWNENEMY] = {
	spawnstate = S_CARRYDISPLAY,
	spawnhealth = 1,
	deathstate = S_THROWNENEMY_SPAWN,
	xdeathstate = S_THROWNENEMY_SPAWN,
	flags = MF_MISSILE
}

local veggie = {
	state = S_VEGGIE,
	sprite = SPR_VEGI,
	frame = A,
	anim_duration = 0,
	renderflags = 0,
	blendmode = AST_TRANSLUCENT,
	spritexscale = FU,
	spriteyscale = FU,
	spritexoffset = 0,
	spriteyoffset = 0,
	radius = 16*FU,
	height = 32*FU,
	tics = 1,
	flags = 0,
	flags2 = 0,
	color = 0,
	colorized = false,
	drawonlyforplayer = nil,
	dontdrawforviewmobj = nil,
	hnext = nil,
	hprev = nil,
	health = 0,
	movedir = 0,
	movecount = 0,
	target = nil,
	reactiontime = 0,
	threshold = 0,
	lastlook = 0,
	spawnpoint = nil,
	tracer = nil,
	friction = 0,
	movefactor = 0,
	fuse = 0,
	scale = FU,
	destscale = FU,
	scalespeed = FU/12,
	extravalue1 = 0,
	extravalue2 = 0,
	cusval = 0,
	cvmem = 0,
	mirrored = false,
	shadowscale = 0,
	dispoffset = 0,
	
	hitfunction = function(i,t)
		P_KillMobj(i)
		return true
	end
}

addHook("ShouldDamage",function(t,i,s,d,dt)
	if not (
		t and t.valid
		and i and i.valid
		and t.skin == "smb2"
	)
	or i.unstandable
		return end
		
	if i.flags & MF_ENEMY
	and t.z - t.momz >= (i.z + i.height) - FU
		i.madesolid = i.flags
		i.madesolidtime = 2
		i.flags = ($|MF_SOLID)&~(MF_SPECIAL|MF_FLOAT)
		i.madesolidtarget = t
		i.madesolidlastpos = {
			x = i.x,
			y = i.y,
			z = i.z
		}
		
		if i.madesolid & MF_NOGRAVITY
			P_TryMove(t,i.x,i.y,i.z+i.height,true)
			t.momz = 0
		end
		return false
	end
end,MT_PLAYER)

addHook("PostThinkFrame",function()
	for p in players.iterate
		p.carrylastbuttons = p.cmd.buttons
		
		if p.mo and p.mo.valid
			if p.mo.state == S_PLAY_CARRYCROUCH
			or p.mo.state == S_PLAY_PLUCK1
			or p.mo.state == S_PLAY_PLUCK2
				p.mo.height = P_GetPlayerSpinHeight(p)
			elseif not P_PlayerShouldUseSpinHeight(p)
				p.mo.height = P_GetPlayerHeight(p)
			end
				
			if p.mo.jumpcharge
			and p.mo.jumpcharge > TICRATE
				p.mo.colorized = not $
				
				if p.mo.colorized
					p.mo.color = SKINCOLOR_WHITE
					p.mo.blendmode = AST_OVERLAY
				else
					p.mo.color = p.skincolor
					p.mo.blendmode = AST_TRANSLUCENT
				end
			else
				p.mo.colorized = false
				p.mo.color = p.skincolor
				p.mo.blendmode = AST_TRANSLUCENT
			end
			
			if p.mo.sprite2 == SPR2_STND
			and p.mo.carrying and p.mo.carrying.valid
				p.mo.sprite2 = SPR2_SKID
			end
		end
	end
	
	for mo in mobjs.iterate()
		if mo.type == MT_CARRYDISPLAY
			local t = mo.target
			if t and t.valid
			and t.carrying and t.carrying.valid
			and t.carrydisplay and t.carrydisplay == mo
				mo.dispoffset = t.dispoffset-1
				local z = t.height
				
				if t.state == S_PLAY_PLUCK1
					z = ease.linear(t.tics*FU/(TICRATE/12),$,0)
					mo.dispoffset = t.dispoffset+1
				end
				
				P_MoveOrigin(mo,t.x,t.y,t.z+z)
			return end
		return end
		
		if mo.flags & (MF_PUSHABLE|MF_MONITOR)
		and not mo.madesolid
			local nearest = nil
			
			searchBlockmap("objects",function(r,f)
				if f.player
				and f.skin == "smb2"
				and R_PointToDist2(r.x,r.y,f.x,f.y) <= r.radius + f.radius
				and f.z <= r.z + r.height + FU
				and f.z > r.z
				and not (nearest and nearest.valid
					and R_PointToDist2(r.x,r.y,nearest.x,nearest.y) <= R_PointToDist2(r.x,r.y,f.x,f.y)
				)
					nearest = f
				end
			end,mo)
			
			if nearest
				mo.madesolid = mo.flags
				mo.madesolidtime = 2
				mo.flags = ($|MF_SOLID)&~(MF_SPECIAL|MF_FLOAT)
				mo.madesolidtarget = nearest
				mo.madesolidlastpos = {
					x = mo.x,
					y = mo.y,
					z = mo.z
				}
				
				mo.momz = 0
				if mo.madesolid & MF_NOGRAVITY
					P_TryMove(nearest,mo.x,mo.y,mo.z+mo.height,true)
					nearest.momz = 0
				end
			end
		end
		
		if mo.madesolidtime
		and mo.madesolidtarget
		and mo.madesolidtarget.skin == "smb2"
			local t = mo.madesolidtarget
			if R_PointToDist2(mo.x,mo.y,t.x,t.y) <= mo.radius + t.radius
			and t.z <= mo.z + mo.height + (32*FU)
			and t.z > mo.z
				t.standingon = mo
				mo.madesolidtime = 2
				
				if mo.madesolid & MF_NOGRAVITY
					mo.momz = min($,-FU/2)
					if not (t.player.pflags & PF_JUMPED)
						t.momz = min($,-FU/2)
					end
				end
				
				local delta = {
					x = mo.x - mo.madesolidlastpos.x,
					y = mo.y - mo.madesolidlastpos.y,
					z = mo.z - mo.madesolidlastpos.z
				}
				
				if delta.z > 0
					mo.z = $-delta.z
				end
				
				P_TryMove(t,t.x+delta.x,t.y+delta.y,mo.z+mo.height,true)
				
				if t.standingon and t.standingon.valid and t.standingon == mo
					P_SpawnLockOn(t.player,mo,S_LOCKON1)
				end
				
				mo.madesolidlastpos = {
					x = mo.x,
					y = mo.y,
					z = mo.z
				}
			else
				mo.madesolidtime = $-1
				t.flags = $&~MF_NOGRAVITY
			end
		elseif mo.madesolid
			mo.flags = mo.madesolid
			mo.madesolidtime = 0
			mo.madesolid = nil
			mo.madesolidtarget = nil
		end
	end
end)

addHook("PreThinkFrame",function()
	for p in players.iterate
		local mo = p.mo
		if not (mo and mo.valid)
			return end
			
		if P_IsObjectOnGround(mo)
			p.powers[pw_strong] = 0
			if p.cmd.buttons & BT_CUSTOM1
				mo.state = S_PLAY_CARRYCROUCH
				p.cmd.forwardmove = 0
				p.cmd.sidemove = 0
				if p.speed > 0
					p.skidnow = 1
				end
				
				if mo.jumpcharge
					if mo.jumpcharge >= TICRATE
						mo.jumpcharge = TICRATE*5
					else
						mo.jumpcharge = $+1
					end
				else
					mo.jumpcharge = 1
				end
			else
				if mo.jumpcharge
					mo.jumpcharge = $-1
				else
					mo.jumpcharge = 0
				end
				
				if mo.state == S_PLAY_CARRYCROUCH
					mo.state = S_PLAY_STND
				end
			end
		else
			if mo.jumpcharge
			and mo.jumpcharge > TICRATE
			and p.pflags & PF_JUMPED
				mo.momz = $*2
				p.powers[pw_strong] = STR_BUST
			end
			
			mo.jumpcharge = 0
		end
	end
end)

addHook("SpinSpecial",function(p)
	if not (p.mo and p.mo.valid and p.mo.skin == "smb2") 
		return end
	
	if not (p.carrylastbuttons != nil
	and ~p.carrylastbuttons & BT_SPIN)
	or p.exiting
	or p.powers[pw_nocontrol]
	or p.powers[pw_carry]
		return end
		
	local mo = p.mo
	
	if mo.carrying and mo.carrying.valid
		throwEnemy(mo,p.actionspd)
		mo.state = S_PLAY_FIRE
		return true
	elseif mo.standingon
	and mo.standingon.valid
	and mo.standingon.madesolidtarget
	and mo.standingon.madesolidtarget == mo
		local t = mo.standingon
		p.powers[pw_nocontrol] = TICRATE/3
		S_StartSound(mo,sfx_pluck)
		mo.state = S_PLAY_PLUCK1
		mo.carrying = {}
		
		mo.momx = $/2
		mo.momy = $/2
		mo.momz = $/2
		
		if not (t.pulledfunction
		and t.pulledfunction(t,mo,mo.carrying))
			mobjToTable(mo.carrying,t)
			
			mo.carrying.type = t.type
			
			if t.madesolid
				mo.carrying.flags = t.madesolid
			end
			
			mo.carrying.valid = true
			t.state = S_NULL
		end
		return true
	else
		S_StartSound(mo,sfx_pluck)
		mo.state = S_PLAY_PLUCK1
		mo.momx = $/2
		mo.momy = $/2
		mo.momz = $/2
		
		mo.carrying = veggie
		mo.carrying.type = MT_NULL
		mo.carrying.valid = true
	end
end)

addHook("PlayerThink",function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.skin == "smb2")
		return end
	
	if mo.carrying
	and (P_PlayerInPain(p)
	or p.pflags & (PF_SPINNING|PF_SLIDING|PF_STASIS)
	or p.powers[pw_carry])
		throwEnemy(mo,p.actionspd)
	end
	
	if mo.carrying
	and mo.carrying.valid
		if mo.carrydisplay
		and mo.carrydisplay.valid
		and mo.carrydisplay.type == MT_CARRYDISPLAY
			local c = mo.carrydisplay
			if mo.carrying.sprite
				c.sprite = mo.carrying.sprite
				c.frame = mo.carrying.frame & FF_FRAMEMASK
				c.renderflags = mo.carrying.renderflags
			end
			c.angle = p.drawangle
		else
			local c = P_SpawnMobj(mo.x,mo.y,mo.z+mo.height,MT_CARRYDISPLAY)
			mo.carrydisplay = c
			c.target = mo
		end
		
		if mo.carrying.type == MT_NULL
			p.normalspeed = 32*FU
		else
			p.normalspeed = 40*FU
		end
		p.runspeed = 1
	else
		p.normalspeed = 28*FU
		p.runspeed = 999*FU
	end
end)

addHook("MobjThinker",function(mo)
	if not (mo and mo.valid)
		return end
		
	local t = mo.target
	if not (t and t.valid
	and t.carrying and t.carrying.valid
	and t.carrydisplay and t.carrydisplay == mo)
		mo.state = S_NULL
	return end
end,MT_CARRYDISPLAY)

addHook("MobjThinker",function(mo)
	if not (mo and mo.valid)
		return end
		
	if not (mo.carrying and mo.carrying.valid)
		mo.state = S_NULL
	return end
	
	mo.sprite = mo.carrying.sprite
	mo.frame = mo.carrying.frame & FF_FRAMEMASK
	mo.renderflags = mo.carrying.renderflags
	mo.height = mo.carrying.height
	if mo.valid
		mo.radius = mo.carrying.radius
	end
end,MT_THROWNENEMY)

addHook("ShouldDamage",function(t,i,s,d,dt)
	if i and i.valid
	and i.type == MT_THROWNENEMY
		if i.carrying.hitfunction
		and i.carrying.hitfunction(i,t)
			return end
			
		local thrown = P_SpawnMobj(i.x,i.y,i.z,MT_THROWNENEMY)
		if not thrown and thrown.valid
			return end
				
		thrown.target = i.target
		thrown.angle = i.angle
		thrown.hitsomething = true
		
		thrown.momx = i.momx/2
		thrown.momy = i.momy/2
		thrown.momz = -i.momz
		
		thrown.carrying = {}
		mobjToTable(thrown.carrying,i.carrying)
		
		thrown.carrying.type = i.carrying.type
		thrown.carrying.valid = true
		
		i.state = S_NULL
	end
end)

local nostand = function(mo) mo.unstandable = true end

addHook("MobjSpawn",nostand,MT_GFZFISH)
addHook("MobjSpawn",nostand,MT_SPINCUSHION)
addHook("MobjSpawn",nostand,MT_MINUS)

addHook("MobjSpawn",function(mo)
	mo.pulledfunction = function(t,i,table)
		P_DamageMobj(t)
		return true
	end
end,MT_GSNAPPER)

addHook("MobjSpawn",function(mo)
	mo.landfunction = function(t,i,hit)
		P_DamageMobj(t)
		
		for fof in t.subsector.sector.ffloors()
			if fof.flags & FOF_BUSTUP
				EV_CrumbleChain(t.subsector.sector,fof)
			end
		end
		
		return true
	end
	
	mo.hitfunction = function(i,t)
		P_KillMobj(i)
		return true
	end
end,MT_TNTBARREL)

addHook("PlayerCanEnterSpinGaps",function(p)
	if p.mo and p.mo.valid
	and p.mo.skin == "smb2"
	and p.mo.state == S_PLAY_CARRYCROUCH
		return true
	end
end)